{-# LANGUAGE QuasiQuotes, OverloadedStrings, FlexibleContexts, UnicodeSyntax,
             EmptyDataDecls, ScopedTypeVariables,
             GeneralizedNewtypeDeriving, FlexibleInstances, RankNTypes #-}
{-# OPTIONS -Wall #-}
module FreshLookHelpers where

import Language.LaTeX
import qualified Language.LaTeX.Builder as B
import qualified Language.LaTeX.Builder.Internal as BI
import qualified Language.LaTeX.Builder.Beamer as BM
import qualified Language.LaTeX.Builder.Math as M
import qualified Language.LaTeX.Length as L
import Language.LaTeX.Builder.QQ
import Control.Monad.Writer
import Control.Applicative
import Data.String
import Data.Monoid.Unicode

p :: LatexItem -> ParItemW
pause :: ParItemW
slide :: LatexItem -> ParItemW -> ParItemW
slideL :: LatexItem
          -> BM.Label
          -> ParItemW
          -> ParItemW
ovFromInt :: Int -> BM.Overlay
alt :: [Int] -> LatexItem -> LatexItem -> LatexItem
altI :: [Int] -> LatexItem -> LatexItem -> LatexItem
margins :: PreambleItem
bfHuge :: LatexItem -> LatexItem
slideCs :: LatexItem -> LatexItem -> ParItemW
slideC :: LatexItem -> ParItemW
slideCB :: LatexItem -> ParItemW
paraC :: LatexItem -> ParItemW
paraCHuge :: LatexItem -> ParItemW
paraCB :: LatexItem -> ParItemW
slideO :: LatexItem -> Writer ParItem a -> ParItemW
mathlift :: (LatexItem -> LatexItem)
            -> MathItem
            -> MathItem
visibleI :: [Int] -> LatexItem -> LatexItem
invisibleI :: [Int] -> LatexItem -> LatexItem
onlyenvI :: [Int] -> ParItem -> ParItem
visibleenvI :: [Int] -> ParItem -> ParItem
visibleIncr :: LatexItem -> LatexItem
myMverb :: String -> LatexItem
vphantom :: LatexItem -> LatexItem
myStrut :: LatexItem
myNewline :: LatexItem
myXchar :: (Char -> LatexItem) -> Char -> LatexItem
hs :: String -> LatexItem
strip :: (a -> Bool) -> [a] -> [a]
altHS :: [String] -> ParItemW
itemize :: Writer [ListItem] () -> Writer ParItem ()
itemizeO :: Writer [ListItem] () -> Writer ParItem ()
item' :: (MonadWriter (f ListItem) m, Applicative f) =>
         LatexItem -> LatexItem -> m ()
ma :: MathItem -> LatexItem
hs0 :: String -> HS a
ths0 :: String -> TermHS a
_eq :: HS (a -> b -> c)
eq :: HS a -> HS b -> HS c
_mem :: HS (a -> b -> c)
(∈) :: HS a -> HS b -> HS c
def :: HS a -> HS a -> Row MathItem
sizeRawTy :: LatexItem
sizeTy :: LatexItem
sizeHS :: LatexItem
unicodesymbols :: PreambleItem
bfLarge :: LatexItem -> LatexItem
ttHuge :: LatexItem -> LatexItem
tthuge :: LatexItem -> LatexItem
ttlarge :: LatexItem -> LatexItem
ttLarge :: LatexItem -> LatexItem
ttLARGE :: LatexItem -> LatexItem
description :: Writer [ListItem] ()
               -> Writer ParItem ()


todo :: a -> a
todo = id
{-# DEPRECATED todo "You have something to do here" #-}

ø :: Monoid m => m
ø = (Data.Monoid.Unicode.∅)

p = put . B.para

pause = p BM.pause

vcenter :: ParItem -> ParItem
vcenter x = B.vfill ⊕ x ⊕ B.vfill

slide title = put . BM.slide title . mapNonEmpty vcenter . execWriter

slideL title lbl bk = BM.frame [] [] [BM.label lbl] title mempty !$? bk

put :: ParItem -> ParItemW
put = tell

ovFromInt = BM.ovFrom . BM.ovInt

alt is x y = BM.only (BM.ovInts is) x ⊕ BM.only [ovFromInt (last is+1)] y

altI = BM.alt . BM.ovInts

margins = BM.setbeamersize (BM.TextMarginLeft (L.cm 0.3))
        ⊕ BM.setbeamersize (BM.TextMarginRight (L.cm 0.3))

bfHuge   = B.decls [B._Huge, B.bfseries]
bfLarge  = B.decls [B._Large, B.bfseries]
ttHuge   = B.decls [B._Huge, B.ttfamily]
tthuge   = B.decls [B.huge, B.ttfamily]
ttlarge  = B.decls [B.large, B.ttfamily]
ttLarge  = B.decls [B._Large, B.ttfamily]
ttLARGE  = B.decls [B._LARGE, B.ttfamily]

slideCs title subtitle =
    slide ø . put . B.center . (⊕subt) . B.para $ title
  where subt = mapNonEmpty ((B.vfill ⊕) . B.para) subtitle

slideC x = slideCs x ø

slideCB = hugebaseskipline . slideC . bfHuge

hugebaseskipline x = do
  p $ BI.rawTex "\\setlength{\\baselineskip}{2.5em}"
  x
  p $ BI.rawTex "\\setlength{\\baselineskip}{1.5em}"

paraC = put . B.center . B.para

paraCHuge = paraC . B.decl B._Huge

paraCB = hugebaseskipline . paraC . bfHuge

slideO title = put . BM.slideO title BM.ovIncr . vcenter . execWriter

mathlift f = M.text . f . ma

visibleI = BM.visible . BM.ovInts
invisibleI = BM.invisible . BM.ovInts

onlyenvI = BM.onlyenv . BM.ovInts
visibleenvI = BM.visibleenv . BM.ovInts

visibleIncr = BM.visible BM.ovIncr

myMverb = B.texttt . (⊕ myStrut) . B.spaceProtector (myXchar (M.mchar B.ttchar))  -- M.verb

vphantom = BI.latexCmdArg "vphantom"

myStrut = vphantom (BI.rawTex "$\\{$")

myNewline = myStrut ⊕ B.newline

--myXchar xchar '·'  = B.makebox (L.ex 1) Centered (BI.rawTex "$\\cdotp$") --(xchar '·')
--myXchar xchar '·'  = B.makebox (L.ex 1) Centered (xchar '·')
myXchar _     '\n' = myNewline
myXchar xchar x
  | x `elem` "·ƛ_" = B.makebox (L.ex 1.22) Centered (xchar x)
  | x `elem` "ø="  = B.ttchar x
  | x `elem` "{}"  = M.mchar B.hchar x
myXchar xchar x    = xchar x

hs = myMverb . strip (=='\n')

strip f = reverse . dropWhile f . reverse . dropWhile f

example :: LatexItem -> ParItemW
example = put . BM.block ø . B.para

exampleHS :: String -> ParItemW
exampleHS = example . hs

altHS = example . mconcat . zipWith (BM.only . BM.ovInts . pure) [1..] . map hs

itemize bk = B.itemize !$? bk
itemizeO bk = BM.itemize BM.ovIncr !$? bk
description bk = B.description !$? bk

item :: MonadWriter [ListItem] m => LatexItem -> m ()
item = tell . pure . B.item . B.para
item' x = tell . pure . B.item' x . B.para

itemD :: (Applicative f, Monoid (f ListItem)) =>
         Writer ParItem () -> Writer (f ListItem) ()
itemD bk = pure . B.item !$? bk

ma = B.math
mh :: HS a -> LatexItem
mh = B.math . unHS
mth :: TermHS a -> LatexItem
mth = B.math . unTermHS

class App repr where
  (·) :: repr (a -> b) -> repr a -> repr b

class Cond repr where
  cond :: repr Bool -> repr a -> repr a -> repr a

newtype HS a = HS { unHS :: MathItem }
  deriving (Eq,Show,Num,IsString,Monoid)

instance App HS where
  HS f · HS x = HS (f ⊕ M.thickspace ⊕ x)

instance Cond HS where
  -- bold the keywords
  cond (HS b) (HS t) (HS e)
    = HS ("if "⊕b⊕" then "⊕t⊕" else "⊕e)

    -- do we want a type constraint here?
{-
instance IsString (HS a) where fromString = HS . fromString

instance Monoid (HS a) where
  mempty = HS mempty
  HS x `mappend` HS y = HS (x `mappend` y)
-}

newtype TermHS a = TermHS { unTermHS :: MathItem }
  deriving (Eq,Show,Num,IsString,Monoid)

instance App TermHS where
  TermHS f · TermHS x = TermHS (f ⊕ M.thickspace ⊕ "·" ⊕ M.thickspace ⊕ x)

liftHS :: (MathItem -> MathItem) -> HS a -> HS b
liftHS f (HS x) = HS (f x)

lift2HS :: (MathItem -> MathItem -> MathItem) -> HS a -> HS b -> HS c
lift2HS f (HS x) (HS y) = HS (f x y)

lift3HS :: (MathItem -> MathItem -> MathItem -> MathItem)
        -> HS a -> HS b -> HS c -> HS d
lift3HS f (HS x) (HS y) (HS z) = HS (f x y z)

liftTHS :: (MathItem -> MathItem) -> TermHS a -> TermHS b
liftTHS f (TermHS x) = TermHS (f x)

(<+>) :: MathItem -> MathItem -> MathItem
x <+> y = x ⊕ M.thickspace ⊕ y

infixr 1 $$
($$) :: LatexItem -> LatexItem -> LatexItem
x $$ y = x ⊕ myNewline ⊕ y

data KSet
type KStar = KSet
data KWorld
data SKind
data Name
class HSRel r
instance HSRel KSet
instance HSRel KWorld
class HSType a
instance HSType Integer
instance (HSType a, HSType b) => HSType (a -> b)
instance (HSType a) => HSType (Maybe a)
instance (HSType a) => HSType [a]
instance HSType ()

agdabin :: (IsString α, Monoid α) => α -> α
agdabin op = "_" ⊕ op ⊕ "_"
binop :: MathItem -> MathItem -> MathItem -> MathItem
binop op x y = x <+> op <+> y
binHS :: HS (a -> b -> c) -> HS a -> HS b -> HS c
binHS = lift3HS binop
ƛ :: HS _Name -> TermHS a -> TermHS b
ƛ (HS x) (TermHS e) = TermHS $ "ƛ" <+> x <+> e
var :: HS _Name -> TermHS a
var (HS x) = TermHS $ "V" <+> x
data BinopPack = B { unapp :: forall a b c. HS (a -> b -> c)
                   , applied :: forall a b c. (HS a -> HS b -> HS c)
                   }
binoppackpoly :: MathItem -> BinopPack
binoppackpoly op = B (HS (agdabin op)) (binHS (HS op))
binoppack :: MathItem -> (HS (a -> b -> c), HS a -> HS b -> HS c)
binoppack op = (HS (agdabin op), binHS (HS op))
hs0 x = HS (M.text (hs x))
ths0 x = TermHS (M.text (hs x))
_Maybe :: HS (KStar -> KStar)
_Maybe = hs0 "Maybe"
_World :: HS KSet -- SKind
_World = hs0 "World"
_Name :: HS (KWorld -> KStar)
_Name = hs0 "Name"
_RawName :: HS KStar
_RawName = hs0 "Name"
_Bool :: HS KStar
_Bool = hs0 "Bool"
_Fin :: HS (Integer -> KStar)
_Fin = hs0 "Fin"
nothing :: HS (Maybe a)
nothing = hs0 "nothing"
just :: HS (a -> Maybe a)
just = hs0 "just"
suc :: HS (Integer -> Integer)
suc = hs0 "suc"
unit :: HS ()
unit = hs0"()"
_Tm :: HS (KWorld -> KStar)
_Tm = hs0 "Tm"
_Set :: HS SKind
_Set = hs0 "Set"
_Binder :: HS (KWorld -> KWorld -> KStar)
_Binder = hs0 "Binder"

class Greek a where
  α :: a
  β :: a
  τ :: a
  σ :: a

instance Greek MathItem where
  α = M.alpha
  β = M.beta
  τ = M.tau
  σ = M.sigma

instance Greek (HS a) where
  α = HS M.alpha
  β = HS M.beta
  τ = HS M.tau
  σ = HS M.sigma

instance Greek (LatexM LatexItm) where
  α = ma M.alpha
  β = ma M.beta
  τ = ma M.tau
  σ = ma M.sigma

sub :: HS a -> HS a
sub = HS . M.sub . unHS
_ℕ :: HS KSet
_ℕ = "ℕ"
_Pred :: HS (KSet -> KSet)
_Pred = hs0 "Pred"
_eqName :: HS Name -> HS Name -> HS Bool
eqNameOp :: MathItem
eqNameOp = "≡" -- ("≟"⊕sub (hs0 "Name"))
eqName_ :: HS (Name -> Name -> Bool)
(eqName_, _eqName) = binoppack eqNameOp
eqNat_ :: HS (Integer -> Integer -> Bool)
eqNat :: HS Integer -> HS Integer -> HS Bool
(eqNat_, eqNat) = binoppack ("≡"{- ≟ -}⊕M.sub "ℕ")
B _eq eq = binoppackpoly "="
B _mem (∈) = binoppackpoly "∈"
(∶) :: HS a -> HS KStar -> HS b
(∶) = binHS ":"
(∶∶) :: HS a -> HS SKind -> HS b
(∶∶) = binHS ":"
-- Here we use a long right arrow
-- to represent a simple right arrow.
(⟶) :: HS KStar -> HS KStar -> HS KStar
(⟶) = binHS "→"
(×) :: HS KStar -> HS KStar -> HS KStar
(×) = binHS "×"
parensHS :: HS a -> HS a
parensHS = liftHS M.parens
parensTHS :: TermHS a -> TermHS a
parensTHS = liftTHS M.parens
eqNameN_ :: HS (Integer -> Integer -> Bool)
eqNameD_ :: HS (Integer -> Integer -> Bool)
eqNameN_ = HS $ agdabin eqNameOp
eqNameD_ = HS $ agdabin eqNameOp
nameOfBinderN :: HS (Integer -> Integer)
nameOfBinderN = hs0 "nameOfBinder"
nameOfBinderD_ :: HS (() -> Integer)
nameOfBinderD_ = hs0 "nameOfBinder"
nameOfBinderD :: () -> HS Integer
nameOfBinderD () = nameOfBinderD_ · unit
exportBinderN :: HS (Integer -> Integer -> Maybe Integer)
exportBinderN = hs0 "exportBinder"
exportBinderD_ :: HS (() -> Integer -> Maybe Integer)
exportBinderD_ = HS "exportBinder"
exportBinderD :: () -> HS Integer -> HS (Maybe Integer)
exportBinderD () x = exportBinderD_ · unit · x
infix 0 `def`
def (HS x) (HS y) = M.cells [x,"≝",y]
sizeRawTy = hs[$str|size : Tm → ℕ|]
sizeTy = hs[$str|size : ∀ {α} → Tm α → ℕ|]
sizeHS =
       hs[$istr|
         |size (V _)        = 1
         |size (t · u)      = 1 + size t + size u
         |size (ƛ _ t)      = 1 + size t
         |size (Let _ t u)  = 1 + size t + size u
         |]


unicodesymbols = BI.rawPreamble [$str|
\providecommand\mathbbm{\mathbb}
\DeclareUnicodeCharacter{737}{\textsuperscript{l}}
\DeclareUnicodeCharacter{8718}{\ensuremath{\blacksquare}}
\DeclareUnicodeCharacter{8759}{::}
\DeclareUnicodeCharacter{9669}{\ensuremath{\triangleleft}}
\DeclareUnicodeCharacter{8799}{\ensuremath{\stackrel{\scriptscriptstyle ?}{=}}}
\DeclareUnicodeCharacter{10214}{\ensuremath{\llbracket}}
\DeclareUnicodeCharacter{10215}{\ensuremath{\rrbracket}}

\DeclareUnicodeCharacter{231}{\c{c}}                            % ç
\DeclareUnicodeCharacter{411}{\textcrlambda}                    % ƛ
\DeclareUnicodeCharacter{945}{\ensuremath{\alpha}}              % α
\DeclareUnicodeCharacter{946}{\ensuremath{\beta}}               % β
\DeclareUnicodeCharacter{947}{\ensuremath{\gamma}}              % γ
\DeclareUnicodeCharacter{948}{\ensuremath{\delta}}              % δ
\DeclareUnicodeCharacter{949}{\ensuremath{\epsilon}}            % ε
\DeclareUnicodeCharacter{953}{\ensuremath{\iota}}               % ι
\DeclareUnicodeCharacter{954}{\ensuremath{\kappa}}              % κ
\DeclareUnicodeCharacter{955}{\ensuremath{\lambda}}             % λ
\DeclareUnicodeCharacter{957}{\ensuremath{\nu}}                 % ν
\DeclareUnicodeCharacter{959}{\ensuremath{o}}                   % ο
\DeclareUnicodeCharacter{963}{\ensuremath{\sigma}}              % σ
\DeclareUnicodeCharacter{964}{\ensuremath{\tau}}                % τ
\DeclareUnicodeCharacter{969}{\ensuremath{\omega}}              % ω
\DeclareUnicodeCharacter{915}{\ensuremath{\Gamma}}              % Γ
\DeclareUnicodeCharacter{916}{\ensuremath{\Delta}}              % Δ
\DeclareUnicodeCharacter{183}{\ensuremath{\cdot}}               % ·
\DeclareUnicodeCharacter{8605}{\ensuremath{\leadsto}}           % ↝
\DeclareUnicodeCharacter{8614}{\ensuremath{\mapsto}}            % ↦
\DeclareUnicodeCharacter{8636}{\ensuremath{\leftharpoonup}}     % ↼
\DeclareUnicodeCharacter{8658}{\ensuremath{\Rightarrow}}        % ⇒
\DeclareUnicodeCharacter{10230}{\ensuremath{\longrightarrow}}   % ⟶
\DeclareUnicodeCharacter{9733}{\ensuremath{\star}}              % ★
\DeclareUnicodeCharacter{9659}{\ensuremath{\vartriangleright}}  % ▻
%\DeclareUnicodeCharacter{9669}{\ensuremath{\vartriangleleft}}  % ◅

\newcommand\defeq{\stackrel{\text{def}}{=}}
\DeclareUnicodeCharacter{8797}{\ensuremath{\defeq}}             % ≝
\newcommand\textLambda{\ensuremath{\Lambda}}
|]
