{-# LANGUAGE QuasiQuotes, OverloadedStrings, FlexibleContexts, UnicodeSyntax,
             GeneralizedNewtypeDeriving, FlexibleInstances #-}
{-# OPTIONS_GHC -F -pgmF frquotes #-}
{-# OPTIONS -Wall -fno-warn-missing-signatures #-}

import Language.LaTeX
import qualified Language.LaTeX.Builder as B
import qualified Language.LaTeX.Builder.Internal as BI
import qualified Language.LaTeX.Builder.Beamer as BM
import qualified Language.LaTeX.Builder.Math as M
import qualified Language.LaTeX.Length as L
import Language.LaTeX.Slicer
import Language.LaTeX.Builder.QQ
import Control.Monad.Writer
import Data.Monoid.Unicode
import FreshLookHelpers

docName = "fresh-look"

main = quickView myViewOpts{basedir="out",showoutput=False,pdfviewer="echo"} docName doc

doc = B.document dc preamb body where
  dc = BM.beamer (Just (L.pt 11)) [BM.t,BM.red,BM.compress] []

usepackages = mconcat . map (BI.usepackage [] . BI.pkgName)

preamb = ø
       ⊕ usepackages ["amsmath","amssymb"]
       ⊕ BI.usepackage ["T3","OT1"] (BI.pkgName "fontenc")
      -- These two declarations are from the tipa package.
      -- However importing the packages breaks the document.
      -- Note that we require T3 fonts here.
       ⊕ BI.rawPreamble [$str|\DeclareTextSymbol\textcrlambda{T3}{172}|]
       ⊕ BI.rawPreamble [$str|\DeclareTextSymbolDefault\textcrlambda{T3}|]
       ⊕ usepackages ["ucs"]
       ⊕ BI.usepackage ["utf8x"] (BI.pkgName "inputenc")
       ⊕ B.title «A fresh look at programming with names and binders»
          -- 'and' improve
       ⊕ B.author «Nicolas Pouillard and François Pottier»
       ⊕ B.institute «INRIA»
       ⊕ BI.rawPreamble "\\date{ICFP 2010, Baltimore\\\\\nSeptember 28, 2010}"
       ⊕ BM.beamertemplatenavigationsymbolsempty
       ⊕ BM.useoutertheme [] "default"
       ⊕ BM.useinnertheme [("shadow","true")] "rounded"
       ⊕ BM.usecolortheme [] "orchid"
       ⊕ footline
       ⊕ margins
       ⊕ unicodesymbols

footline :: PreambleItem
       -- ⊕ BM.footline BM.defaultFootline{BM.showTotalFrames=False}
footline =
  BI.rawPreamble $
  [$istr|\defbeamertemplate*{footline}{}
        |{
        |  \leavevmode%
        |  \hbox{%
        |    \begin{beamercolorbox}[wd=\paperwidth,ht=2.25ex,dp=1.125ex,right]{}%
        |      \insertframenumber{}\hspace*{2ex}
        |    \end{beamercolorbox}}%
        |  \vskip0pt%
        |}
        |]

body = slice . execWriter $ do
  put B.maketitle

  slideCB «Program representation should stay well-typed and {BM.alert «well-scoped»}»
  -- In order to apply the same kind of rules to the programs we manipulate
  -- than those we apply to the program that manipulates them.
  --
  -- By meta-programming we means all programs that manipulate some kind of programs
  --
  -- the well-typed part guarantee that we can only build well-typed program
  --   representation, the benefit counter part is that you learn part of the
  --   type when pattern matching a program
  -- here we will omit the "well-typed" part since this talk is dedicated to
  --   bindings and scopes
  -- the well-scoped part guarantee that we don't generate programs with
  --   "unbound values", or "free variables"

  slideCB «What do we need to manipulate data structures with names and binders?»
  -- ORAL: situation w.r.t. related work!
  -- PLAN: I will start with vastely common way to manipulate ... and then
  --       I will present our improvement on that.
  slide ø $ do
    p              «A type for names»
    example . mh $ _Name ∶∶ _Set
    pause
    p              «An equality test for names»
    example . mh $ eqName_ ∶ (_RawName ⟶ _RawName ⟶ _Bool)
    pause
    p              «An infinite set of distinct names (gensym, stream, whatever)»
    exampleHS      [$str|x, y, z, … : Name|]

  BM.frame (BM.ovInts [2]) ø [BM.label "dataTm"]
           «Example: Data type for the untyped lambda calculus» ø . vcenter !$?
    altHS
      [ [$istr|
        |data Tm
        | =  V    Name
        | |  _·_  Tm Tm
        | |  ƛ    Name Tm
        | |  Let  Name Tm Tm
        |]
      , [$istr|
        |data Tm : Set where
        | V    : Name → Tm
        | _·_  : Tm → Tm → Tm
        | ƛ    : Name → Tm → Tm
        | Let  : Name → Tm → Tm → Tm
        |]
      , [$istr|
        |data Tm : Set where
        | V    : (x : Name) → Tm
        | _·_  : (t u : Tm) → Tm
        | ƛ    : (x : Name) (t : Tm) → Tm
        | Let  : (x : Name) (t u : Tm) → Tm
        |]
{-
      , [$istr|
        |data Tm outer
        | =          V    (Name outer)
        | |          _·_  (Tm outer)    (Tm outer)
        | | ∃inner.  ƛ    (Binder outer inner)  (Tm inner)
        | | ∃inner.  Let  (Binder outer inner)  (Tm outer) (Tm inner)
        |]
-}
      , [$istr|
        |data Tm α
        | =        V    (Name α)
        | |        _·_  (Tm α) (Tm α)
        | |  ∃ β.  ƛ    (Binder α β)  (Tm β)
        | |  ∃ β.  Let  (Binder α β)  (Tm α) (Tm β)
        |]
      , [$istr|
        |data Tm α : Set where
        | V    : Name α → Tm α
        | _·_  : Tm α → Tm α → Tm α
        | ƛ    : ∀{β} → Binder α β → Tm β → Tm α
        | Let  : ∀{β} → Binder α β → Tm α → Tm β → Tm α
        |]
      , [$istr|
        |data Tm α : Set where
        | V    : Name α → Tm α
        | _·_  : Tm α → Tm α → Tm α
        | ƛ    : Abs Tm α → Tm α
        | Let  : Abs (λ β → Tm α × Tm β) α → Tm α
        |]
      , [$istr|
        |data Tm α : Set where
        | V    : (x : Name α) → Tm α
        | _·_  : (t u : Tm α) → Tm α
        | ƛ    : ∀{β} (x : Binder α β) (t : Tm β) → Tm α
        | Let  : ∀{β} (x : Binder α β) (t : Tm α) (u : Tm β) → Tm α
        |]
      ]

  slide «Examples using this data type» $ do
    p         «Building terms»
    exampleHS [$istr|x, y :: Name
                    |idTerm, apTerm :: Tm
                    |idTerm = ƛ x (V x)
                    |apTerm = ƛ x (ƛ y (V x · V y))
                    |]
    p         «Analysing terms»
    example $ sizeRawTy
           $$ sizeHS

  slideCB $ alt [1] «Nothing more!» «Something less?»
  -- nothing more! the programming model is quite expressive
  -- something less? maybe there is functions and terms we
  -- would like not to be able to create?
  slide ø $ do
    let x = hs0 "x"
        y = hs0 "y"
        e = ths0 "e"
    paraCHuge . mth $ ƛ x e · var y
    paraCB            «Shall we compare {mh x} and {mh y}?
                       {visibleI [2] $ BM.alert «No!»}»

  slideCB «Thou Shalt Not Discriminate α-equivalent terms!»
  slide ø $ do
    paraCB         «Comparison is World Homogeneous»
    example . mh $ eqName_ ∶ hs0[$str|∀ {α} → Name α → Name α → Bool|]

  slide ø $ do
    paraCB    «Names inhabit worlds»
    exampleHS [$istr|World  : Set
                    |Name   : World → Set
                    |ø      : World
                    |]
    p         «Worlds will be either ø or abstract variables (like α,β)»

  slide ø $ do
    let wa = (⊕(HS . mathlift BM.alert $ M.sub α))
        wb = (⊕(HS . mathlift BM.alert $ M.sub β))
        x  = hs0 "x"
        y  = hs0 "y"
        z  = hs0 "z"
    paraCHuge . visibleIncr . mth $
      parensTHS (ƛ (wb x) (parensTHS . var . wb $ z)) · var (wa y)
    itemizeO $ do
      item «Comparing {mh x} and {mh y} is now forbidden!»
      item «Comparing {mh x} and {mh z} is allowed.»
      item «One could expect to be able to compare {mh z} and {mh y}
            when {mh x} and {mh z} are distinct!»
    paraCB . visibleIncr $ «We need links between worlds!»

  slide «A type for binders» $ do
    paraCB . mh $ _Binder·α·β ∶∶ _Set
    itemize $ do
      item «A more precise type for a Name β»
      item «Where α is the outer world» -- before the binding
      item «And β is the inner world» -- after the binding
    exampleHS [$str|nameOfBinder : ∀ {α β} → Binder α β → Name β|]

  put $ BM.againframe (BM.ovInts [2,5]) ø [] "dataTm"

  slide ø $ do
    paraCB «Name abstraction is existential quantification over scopes»
    exampleHS [$istr|Abs : (World → Set) → World → Set
                    |Abs F α = ∃ β, Binder α β × F β|]
  put $ BM.againframe (BM.ovInts [6]) ø [] "dataTm"

  slideO «Computing the size of a term» $ do
    example $ altI [1] sizeRawTy sizeTy $$ sizeHS
    pause
    itemizeO $ mapM_ item
      [ «Polymorphic recursion at work!»
      , «No freshening! (neither explicit nor implicit)»
      , «No shifting!»
      ]

  slide «Exporting a name through a binder» $ do
    exampleHS
      [$istr|exportBinder :  ∀ {α β} (x : Binder α β) (y : Name β)
            |                → Maybe (Name α)|]
    let y       = hs0 "y"
        x       = hs0 "x"
        nOfx    = nameOfBinderN·x
    itemizeO $ do
      item «Returns {mh (just·y)} when {mh y} and {mh nOfx} are distinct»
      item «Performs a type refinement on success»
      item «Partial!»

  slide «Computing free variables» $ do
    example . visibleI [2..4] $
      altI [2,3] (hs
      [$istr|
      |rm : Name → List Name → List Name
      |rm _  []                   = []
      |rm x  (y ∷ ys)  with x ≡ y
      |...             | false    = y ∷ rm x ys
      |...             | true     = rm x ys
      |])
      (hs
      [$istr|
      |rm : ∀ {α β} → Binder α β → List (Name β) → List (Name α)
      |rm _  []                   = []
      |rm x  (y ∷ ys)  with exportBinder x y
      |...             | just y   = y ∷ rm x ys
      |...             | nothing  = rm x ys
      |])
    example . visibleI [1..4] $
      altI [1,2] (hs[$str|fv : Tm → List Name|])
                 (hs[$str|fv : ∀ {α} → Tm α → List (Name α)|])
      $$hs
      [$istr|fv (V x)        = [ x ]
            |fv (t · u)      = fv t ++ fv u
            |fv (ƛ x t)      = rm x (fv t)
            |fv (Let x t u)  = fv t ++ rm x (fv u)
            |]

  slide «Two implementations» $ do
    let eqns = B.displaymath . M.array [B.r,B.c,B.l]
    onlyenvI [2] !$? do
      p «Nominal» 
      let a, b :: HS Integer
          (a,b) = ("a","b")--(M.a,M.b)
      eqns !$
        [_World             `def` _Pred·_ℕ
        ,_Name·β            `def` "{ a : ℕ | "⊕a ∈ β⊕" }"
        ,_Binder·α·β        `def` "{ a : ℕ | β = {a} ∪ α }"
        ,eqNameN_           `def` eqNat_
        ,exportBinderN·a·b  `def` cond (eqNat a b) nothing (just · b)
        ,nameOfBinderN·a    `def` a
        ]

    onlyenvI [3] !$? do
      p «De Bruijn indices»
      let i = HS M.i
          ie = M.text «(i.e. {B.math (mconcat [M.mleft '[',0,",",β,M.mright ')'])})»
      eqns !$
        [_World                 `def` _ℕ
        ,_Name·β                `def` _Fin·β
        ,M.cells [ø,ø,ie]
        ,_Binder·α·β            `def` (β `eq` (α + 1))
        ,eqNameD_               `def` eqNat_
        ,exportBinderD () 0     `def` nothing
        ,exportBinderD ()
          (parensHS (suc · i))  `def` just · i
        ,nameOfBinderD ()       `def` 0
        ]

  slideCB «Well-typed programs preserve α-equivalence!»

  slide «What do we mean by α-equivalence?» . itemize $ do
    item «One builds a logical relation indexed by types»
    item «It models observational equivalence between programs»
    item «It coincides with traditional α-equivalence at type {mh _Tm} (for instance)»
    item «It extends α-equivalence at richer types»

  slide «Free theorems» $ do
    let f = hs0 "f"
        x = hs0 "x"
    exampleHS [$str|f : ∀ {α} → Tm α → Tm α|]
    itemize $ do
      item «Every name that occurs free in {mh(f·x)} must occur free in {mh x}»
      item «Furthermore, {mh f} commutes with any renaming»
      item «Using well-scoped De Bruijn indices would not yield the latter guarantee»

{-
  slide «Three views of Worlds» $ do
    itemize $ do
      item «Nothing (a phantom type) since they do not exist at runtime»
      item «A set of names»
      item «A partial bijection between names»

  slide «Logical relation internals» $ do
    itemize $ do
      item «Worlds are seen as bijections between names»
      item «A relation is given for primitive types»
      item «Primitive operations are checked to fit the relation»
      item «From this every well-typed program fits the relation»
-}

  slide ø $ do
    p «The paper:»
    itemize $ do
      item «describes more primitive machinery to safely import or generate names»
      item «shows how to lift operations on names (equality test, import, export...) to types like {mh _Tm}»
      item «offers an advanced example (NBE)»
      itemD $ do
        p         «explores a HOAS-like form of abstraction:»
        exampleHS [$str|FunAbs F α = ∀ {β} → α ⊆ β → F β → F β|]

{-
  slide «Challenges and future work» $
    itemize $ do
      item «Reasoning about programs independently of the implementation scheme (nominal or de Bruijn)»
      item «How to carry information on binders (types, specs...)»
      item «More formalisation of the logical relations»
      item «More advanced usages»
-}

  slide «Conclusion» $ do
    itemize $ do
      item «Explicit scopes using world indices»
      item «No freshening when pattern-matching abstractions»
      item «Name abstraction as existential quantification»
      item «Expressiveness close to a manual model with names»
      item «Two concrete and safe implementations of the interface»
    p «Future Work: How to reason about programs using this model»

  put BM.appendix

  let go trg titl = B.para $ BM.hyperlink [] trg (BM.beamerbutton titl) []
      gobackbonus =
         B.para !$ BM.hyperlink [] "bonus" (BM.beamerreturnbutton «Back») []
      slideB x y bk = slideL x y (bk >> gobackbonus)
  slideL «Bonus tracks» "bonus" $ do
    go "subst" !$ «Capture avoiding substitution»
    go "occurs" . hs !$ "occurs"
    go "star" . hs !$ "Star"
    go "env" . hs !$ "★Binder"
    go "exportStar" . hs !$ "export★Binder"
    go "fv'" . hs !$ "fv'"
    go "dbcanon" !$ «De Bruijn indices canonicity: precisely»
    go "pack" !$ «Packaging»
    go "app" . hs !$ "app"
    go "context" !$ «Contexts»
    go "patt" !$ «Patterns»
    go "nameEqH" . hs !$ "nameEqH"

  -- put $ BM.againframe ø ø [] "dataTm"
  put $ BM.againframe (BM.ovInts [1..7]) ø [] "dataTm"

  -- slideCB «Polymorphic values represent closed terms»
  -- * forall alpha. Tm alpha
  --   there cannot be any free atom in it
  -- * forall alpha. Tm alpha -> Tm alpha OR Name alpha
  --   free atoms of the output are necessary a subset of input free atoms

  -- slideCB «A more direct presentation of atom sorts»
  -- atom sorts are typically added afterward for pragmatic
  -- reasons, however they add some complexity to the presentation
  -- without adding any theoretical benefit.

  -- slideCB «Generalizing {alphaCaml} data structures»
  -- * alphaCaml has two different kind of types,
  --   pattern types and expression types.
  -- * Pattern types describe the shape of binders.
  -- * Abstraction pack pattern types and expression
  --   types as a single abstract representation
  --   supporting alpha equivalence
  -- * These distinctions where guided by practical
  --   use cases rather than some uniform presentation.

  {-
  slideCB «Reaching {ott} expressiveness and more»
  -- * Ott enables to define the set of bound atoms,
  --   and also where they are bound in a very general way.
  -- * This gives a more general presentation that patt/expr types
  -- * One would like to go further by allowing to dynamically
  --   select bound atoms.
  -}

  slideB «Capture avoiding substitution» "subst" $ do
    exampleHS $
      [$istr|
|module Substitution where
|  Fresh×Name→F : (F : World → Set) (α β : World) → Set
|  Fresh×Name→F F α β = Fresh β × (Name α → F β)
|
|  commθ : ∀ {F} → Import⊆ F → (∀ {α} → Name α → F α)
|                → Comm (Fresh×Name→F F)
|  commθ imp var x (y , f)
|    = (_ , (nextOf y , f' ∘ exportBinder x) , weakOf y)
|    where
|      open FreshPack
|      f' = maybe (imp (⊆Of y) ∘ f) (var (nameOf y))
|]

  slide «Capture avoiding substitution (part 2)» $ do
    exampleHS $
      [$istr|
|open Substitution
|substTm : ∀ {α β} → Fresh×Name→F Tm α β → Tm α → Tm β
|substTm θ (V x) = proj₂ θ x
|substTm θ (t · u)
|  = substTm θ t · substTm θ u
|substTm θ (ƛ x t)
|    with commθ impTm V x θ
|... | (_ , θ' , x') = ƛ x' (substTm θ' t)
|substTm θ (Let x t u)
|    with commθ impTm V x θ
|... | (_ , θ' , x') = Let x' (substTm θ t) (substTm θ' u)
|]


  slideB «Occurs check» "occurs" $
    exampleHS
      [$istr|
      |occurs : ∀ {α} → Name α → Tm α → Bool
      |occurs x₀ = occ (λ y → x₀ ≡ y)
      |  where
      |    OccEnv  : World → Set
      |    OccEnv  α = Name α → Bool
      |    extend  : ∀ {α β} → Binder α β → OccEnv α → OccEnv β
      |    extend  x Γ y = maybe Γ false (exportBinder x y)
      |
      |    occ : ∀ {α} → OccEnv α → Tm α → Bool
      |    occ Γ  (V y)        = Γ y
      |    occ Γ  (t · u)      = occ Γ t ∨ occ Γ u
      |    occ Γ  (ƛ x t)      = occ (extend x Γ) t
      |    occ Γ  (Let x t u)  = occ Γ t ∨ occ (extend x Γ) u
      |]

  slideB «Reflexive and transitive closure of a relation» "star" $ do
    exampleHS
      [$str|Rel A = A → A → Set|]
    exampleHS
      [$istr|
      |data Star {I : Set} (T : Rel I) : Rel I where
      |  ε    : ∀ {i} → Star T i i
      |  _◅_  : ∀ {i j k} → T i j → Star T j k → Star T i k
      |]
    exampleHS
      [$istr|
      |Star⁻¹  : ∀ {I : Set} → Rel I → Rel I
      |Star⁻¹  = flip ∘ Star ∘ flip
      |]

  slideB «Environments as a chain of binders» "env" $ do
    exampleHS
      [$istr|
      |★Binder  : Rel World
      |★Binder  = Star⁻¹ Binder
      |]
    exampleHS
      [$istr|
      |  ε  : ∀ {α} → ★Binder α α
      |_◅_  : ∀ {α β γ}  (x : Binder β γ) (Γ : ★Binder α β)
      |               →  ★Binder α γ
      |]

  slideB «Exporting through a chain of binders» "exportStar" $
    exampleHS
      [$istr|
      |export★Binder  : ∀ {α β}  → ★Binder α β
      |                          → Name β → Maybe (Name α)
      |export★Binder  ε        y
      |  = just y
      |export★Binder  (x ◅ Γ)  y
      |  = exportBinder x y >>= export★Binder Γ
      |  where open MaybeMonad
      |]

  slideB «Free variables, continued...» "fv'" $
    exampleHS
      [$istr|
      |fv'  : ∀ {β α} → ★Binder α β → Tm β → List (Name α)
      |fv'  Γ  (V x)        = List.fromMaybe (export★Binder Γ x)
      |fv'  Γ  (t · u)      = fv' Γ t ++ fv' Γ u
      |fv'  Γ  (ƛ x t)      = fv' (x ◅ Γ) t
      |fv'  Γ  (Let x t u)  = fv' Γ t ++ fv' (x ◅ Γ) u
      |]

  slideL «De Bruijn indices canonicity: precisely» "dbcanon" . itemize $ do
    item «Bound and fresh names uniquely chosen.»
    item «Free names are freely chosen, though!»
    item «Consequence: α-equivalence is not so simple!»
    item «α-equivalences are not automatically preserved!»

  slideB «Packaging up...» "pack" . itemize $ do
    let itm = item . hs
    itm "binderOf"
    itm "nameOf"
    itm "nextOf"
    itm "importWith"

  slideB «Constructing terms» "app" $
    exampleHS
      [$istr|
      |app  : Tm ø
      |app  = ƛ  (binderOf x) (ƛ (binderOf y)
      |          (V (importWith y (nameOf x)) · V (nameOf y)))
      |  where  open FreshPack
      |         x  = freshø
      |         y  = nextOf x
      |]

  slideB «Towards elaborate uses of worlds» "context" $
    exampleHS
      [$istr|
      |data C α : (β : World) → Set where
      |  Hole  : C α α
      |  _·₁_  : ∀ {β} (c : C α β) (t : Tm α) → C α β
      |  _·₂_  : ∀ {β} (t : Tm α) (c : C α β) → C α β
      |  ƛ     : ∀ {β γ}  (x : Binder α β) (τ : Ty)
      |                   (c : C β γ) → C α γ
      |  Let₁  : ∀ {β γ}  (x : Binder α β) (c : C α γ)
      |                   (t : Tm β) → C α γ
      |  Let₂  : ∀ {β γ}  (x : Binder α β) (t : Tm α)
      |                   (c : C β γ) → C α γ
      |]

  slideB «Patterns types» "patt" $
    exampleHS
      [$istr|
      |data Pa α β : Set where
      |  PVar   : Binder α β → Pa α β
      |  PPair  : ∀{γ} → Pa α γ → Pa γ β → Pa α β
      |]

  slideB «Safe heterogeneous comparison!» "nameEqH" $ do
    exampleHS [$str|nameEqH : ∀{α β}→ Name α → Name β → Binder α β → Bool|]
    example $ hs [$istr|nameEqH x y lnk with exportBinder lnk y
                       |...                | just y' = x ≡ y'
                       |...                | nothing = false|]

  -- slideL «What about moving names the other way around? {visibleI [2] $ BM.alert «No!»}»
  -- TODO
